package cz.muni.fi.pb138.fsmda.comparison;

import java.util.*;
import java.util.logging.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Implementation of FileSystemComparer. Uses BFS.
 *
 * @author Matus Nemec
 */
public class FileSystemComparerImpl implements FileSystemComparer {

    static final Logger logger = Logger.getLogger(ComparisonManagerImpl.class.getName());

    private void fillMaps(Map<String, Node> dirs, Map<String, Node> files, Node node) {
        if (node != null) {
            NodeList children = node.getChildNodes();
            for (int i = 0; i < children.getLength(); i++) {
                if ("file".equals(children.item(i).getNodeName())) {
                    files.put(((Element) children.item(i)).getAttribute("name"),
                            children.item(i));
                } else if ("directory".equals(children.item(i).getNodeName())) {
                    dirs.put(((Element) children.item(i)).getAttribute("name"),
                            children.item(i));
                }
            }
        }
    }
    List<String> differences = new ArrayList<String>();

    private List<String> compareItems(Node control, Node test, String fullPath, boolean file) {
        Element controlEl = (Element) control;
        Element testEl = (Element) test;
        List<String> diffs = new ArrayList<String>();
        String type;
        String typeSingular;
        if (file) {
            type = "files ";
            typeSingular = "file";
        } else {
            type = "directories ";
            typeSingular = "directory";
        }
        if (!controlEl.getAttribute("length").equals(testEl.getAttribute("length"))) {
            diffs.add("Size of " + type + "\"" + fullPath + "\" is not equal.\n"
                    + "    Size in control document is " + controlEl.getAttribute("length") + " bytes."
                    + "\n    Size in test document is " + testEl.getAttribute("length") + " bytes.");
        }
        if (!controlEl.getAttribute("lastModified").equals(testEl.getAttribute("lastModified"))) {
            diffs.add("Time of last modification of " + type + "\"" + fullPath + "\" is not equal.\n"
                    + "    Last modification in control document was " + controlEl.getAttribute("lastModified")
                    + "\n    Last modification in test document was " + testEl.getAttribute("lastModified"));
        }
        String params[] = {"executable", "readable", "writable"};
        for (String par : params) {
            if (!controlEl.getAttribute(par).equals(testEl.getAttribute(par))) {
                diffs.add("Value of parameter " + par + " in " + type + "\"" + fullPath + "\" is not equal.\n"
                        + "    It is set to " + controlEl.getAttribute(par) + " in control document and "
                        + "to " + testEl.getAttribute(par) + " in test document.");
            }
        }
        if ("true".equals(controlEl.getAttribute("hidden")) 
                && !"true".equals(testEl.getAttribute("hidden"))) {
            diffs.add("Hidden " + typeSingular + " \"" + fullPath + "\" is not hidden in the test document.\n"
                    + "    It was hidden in the control document.");
        }
        if ("true".equals(testEl.getAttribute("hidden"))
                && !"true".equals(controlEl.getAttribute("hidden"))) {
            diffs.add("Not hidden " + typeSingular + " \"" + fullPath + "\" is now hidden in the test document.\n"
                    + "    It was not hidden in the control document.");
        }      
        return diffs;
    }

    private void compareDocuments(Document controlDoc, Document testDoc,
            Node controlRootNode, Node testRootNode, String path) {
        Map<String, Node> controlFiles = new HashMap<String, Node>();
        Map<String, Node> testFiles = new HashMap<String, Node>();
        Map<String, Node> controlDirs = new HashMap<String, Node>();
        Map<String, Node> testDirs = new HashMap<String, Node>();

        fillMaps(controlDirs, controlFiles, controlRootNode);
        fillMaps(testDirs, testFiles, testRootNode);

        for (String name : controlFiles.keySet()) {
            if (testFiles.containsKey(name)) {
                differences.addAll(compareItems(controlFiles.get(name),
                        testFiles.get(name), path + name, true));

                testFiles.remove(name);
            } else {
                differences.add("Missing file \"" + path + name 
                        + "\".\n    There is no such file in the test document.");       
            }
        }

        for (String name : testFiles.keySet()) {
            differences.add("New file \"" + path + name 
                    + "\".\n    There is no such file in the control document.");
        }

        for (String name : controlDirs.keySet()) {
            if (testDirs.containsKey(name)) {
                differences.addAll(compareItems(controlDirs.get(name),
                        testDirs.get(name), path + name, false));

                compareDocuments(controlDoc, testDoc, controlDirs.get(name),
                        testDirs.get(name), path.concat(name + "/"));

                testDirs.remove(name);
            } else {
                differences.add("Missing directory \"" + path + name 
                    + "\".\n    There is no such directory in the test document.");
            }
        }

        for (String name : testDirs.keySet()) {
            differences.add("New directory \"" + path + name 
                    + "\".\n    There is no such directory in the control document.");
        }
    }

    /**
     * Compares filesystem representations using BFS.
     */
    @Override
    public List<String> compareFileSystemRepresentations(Document controlDoc, Document testDoc) {
        logger.info("Start of comparison");
        if (controlDoc == null) {
            logger.info("Trying to compare with null");
            throw new IllegalArgumentException("controlDoc is null");
        }
        if (testDoc == null) {
            logger.info("Trying to compare with null");
            throw new IllegalArgumentException("testDoc is null");
        }

        if (!((Element) controlDoc.getFirstChild()).getAttribute("length").equals(((Element) testDoc.getFirstChild()).getAttribute("length"))) {
            differences.add("Size of the root directory is not equal.\n"
                    + "    Size of control directory is "
                    + ((Element) controlDoc.getFirstChild()).getAttribute("length") + " bytes."
                    + "\n    and size of test directory is "
                    + ((Element) testDoc.getFirstChild()).getAttribute("length") + " bytes.");
        }
        compareDocuments(controlDoc, testDoc, controlDoc.getFirstChild(),
                testDoc.getFirstChild(), "/");

        logger.info("Finnishing comparison");
        return Collections.unmodifiableList(differences);
    }
}
